@@ -8,10 +8,8 @@ namespace miic {
 namespace computation {
 
 structure::InfoBlock getCondMutualInfo(int X, int Y,
-    const std::vector<int>& ui_list,
-    const std::vector<std::vector<int>>& data_numeric,
-    const std::vector<std::vector<int>>& data_numeric_idx,
-    structure::Environment&);
+    const std::vector<int>& ui_list, const structure::Grid2d<int>& data_numeric,
+    const structure::Grid2d<int>& data_numeric_idx, structure::Environment&);
 double getInfo3PointOrScore(structure::Environment&, int X, int Y, int Z,
     const std::vector<int>& ui_list, bool get_info);
 void searchForBestContributingNode(
